unit __PS_rus;

interface

//    
//Porter Stemming
// :
//   http://saxmms.blogspot.com/2010/12/delphi.html

uses SysUtils, Classes, StrUtils;


const
  //   :
  PERFECTIVEGROUND_I = ',,,,,';
  PERFECTIVEGROUND_A = ',,,,,';

  // :
  ADJECTIVE = ',,,,,,,,,,,,,,,,' +
              ',,,,,,,,,';

  // :
  PARTICIPLE_I = ',,';
  PARTICIPLE_A = ',,,,,,,,,';

  // 
  VERB_I = ',,,,,,,,,,,,,,,,' +
           ',,,,,,,,,,,,';
  VERB_A = ',,,,,,,,,,,,,,,' +
           ',,,,,,,,,,,,,,,' +
           ',,,';

  // :
  NOUN = ',,,,,,,,,,,,,,,,,,,,,' +
         ',,,,,,,,,,,,,,';

  VOWEL = '';


function StemWord_Porter(const AWord: String): String;
function StemWordList_Porter(List:TStrings): integer; overload;
function StemWordList_Porter(ListIn:TStrings; ListOut:TStrings): integer; overload;


implementation

var
  DICT_PERFECTIVEGROUND_I,
  DICT_PERFECTIVEGROUND_A,
  DICT_ADJECTIVE,
  DICT_PARTICIPLE_I,
  DICT_PARTICIPLE_A,
  DICT_VERB_I,
  DICT_VERB_A,
  DICT_NOUN: TStringList;

function IsVowel(const AChar: Char): Boolean;
begin
  Result := Pos(AChar, VOWEL) > 0;
end;

function HasTail(const RV: String; Tail: String): Boolean;
begin
  Result :=  RightStr(RV, Length(Tail)) = Tail;
end;

function CutTail(var RV: String; Tail: String): Boolean;
begin
  Result := False;

  if Length(RV) = 0 then Exit;
  if Length(Tail) = 0 then Exit;
  if Length(Tail) > Length(RV) then Exit;

  if RightStr(RV, Length(Tail)) = Tail then
  begin
    Result := True;
    Delete(RV, Length(RV) + 1 - Length(Tail), Length(Tail))
  end;
end;

function ReplaceSuffix(var RV: String; SufList: TStrings;
  LessOne: Boolean): Boolean;
var
  I: Integer;
  Tail: String;
begin
  Result := False;

  for I := 0 to SufList.Count - 1 do
  begin
    Tail := SufList[I];
    if HasTail(RV, Tail) then
    begin
      if LessOne then
        Delete(RV, Length(RV) + 2 - Length(Tail), Length(Tail) - 1)
      else
        Delete(RV, Length(RV) + 1 - Length(Tail), Length(Tail));

      Result := True;
      Exit;
    end;
  end;
end;

function ReplacePERFECTIVEGROUND(var RV: String): Boolean;
begin
  Result := ReplaceSuffix(RV, DICT_PERFECTIVEGROUND_I, False);
  if not Result then
    Result := ReplaceSuffix(RV, DICT_PERFECTIVEGROUND_A, True);
end;

function ReplaceREFLEXIVE(var RV: String): Boolean;
begin
  Result := CutTail(RV, '');
  if not Result then
    Result := CutTail(RV, '');
end;

function ReplaceADJECTIVE(var RV: String): Boolean;
begin
  Result := ReplaceSuffix(RV, DICT_ADJECTIVE, False);;
end;

function ReplacePARTICIPLE(var RV: String): Boolean;
begin
  Result := ReplaceSuffix(RV, DICT_PARTICIPLE_I, False);
  if not Result then
    Result := ReplaceSuffix(RV, DICT_PARTICIPLE_A, True);
end;

function ReplaceVERB(var RV: String): Boolean;
begin
  Result := ReplaceSuffix(RV, DICT_VERB_I, False);
  if not Result then
    Result := ReplaceSuffix(RV, DICT_VERB_A, True);
end;

function ReplaceNOUN(var RV: String): Boolean;
begin
  Result := ReplaceSuffix(RV, DICT_NOUN, False);
end;

procedure StemStep1(var RV: String);
begin
  if not ReplacePERFECTIVEGROUND(RV) then
  begin
    ReplaceREFLEXIVE(RV);

    if ReplaceADJECTIVE(RV) then
      ReplacePARTICIPLE(RV)
    else
    if not ReplaceVERB(RV) then
      ReplaceNOUN(RV);
  end;
end;

procedure StemStep2(var RV: String);
begin
  CutTail(RV, '');
end;

procedure StemStep3(var RV: String);
var
  Temp: String;

  function IsRoot(const Value: String): Boolean;
  var
    State: (st1d, st2d, st3d);
    I: Integer;
  begin
    Result := False;
    if Length(Value) < 2 then Exit;

    State := st1d;
    for I := 2 to Length(Value) do
      case State of
        st1d:
          if IsVowel(Value[I]) then
            State := st2d;

        st2d:
          if not IsVowel(Value[I]) then
            State := st3d;

        st3d:
          Break;
      end;

    Result := State = st3d;
  end;

begin
  Temp := RV;

  if CutTail(Temp, '') and IsRoot(Temp) then
    RV := Temp
  else
  if CutTail(Temp, '') and IsRoot(Temp) then
    RV := Temp;
end;

procedure StemStep4(var RV: String);
begin
  if not CutTail(RV, '') then
  begin
    if not CutTail(RV, '') then CutTail(RV, '');
    if CutTail(RV, '') then RV := RV + '';
  end;
end;

function StemWordList_Porter(List:TStrings): integer;
Var
 Sx:string;
 i:integer;
begin
  Result:=0;
  if Assigned(List) then begin
     i:=-1;
     while i<(List.Count-1) do
      begin
       i:=i+1;
       Sx:=trim(List[i]);
       if length(Sx)>0 then begin
          Sx:=trim(StemWord_Porter(Sx));
          if length(Sx)>0 then begin
             List[i]:=Sx;
             Result:=Result+1;
          end;
       end;
     end;
  end;
end;


function StemWordList_Porter(ListIn:TStrings; ListOut:TStrings): integer;
Var
 Sx:string;
 i:integer;
begin
  Result:=0;
  if Assigned(ListOut) then begin
     ListOut.Clear;
     i:=-1;
     while i<(ListIn.Count-1) do
      begin
       i:=i+1;
       Sx:=trim(ListIn[i]);
       if length(Sx)>0 then begin
          Sx:=trim(StemWord_Porter(Sx));
          if length(Sx)>0 then begin
             ListOut.Add(Sx);
          end;
       end;
     end;
     Result:=ListOut.Count;
  end;
end;

function StemWord_Porter(const AWord: String): String;
var
  Prefix, RV: String;
  I, Index: Integer;
begin
  {$IFDEF UNICODE}
  Result := WideLowerCase(AWord);
  {$ELSE}
  Result := LowerCase(AWord);
  {$ENDIF}
  Result := StringReplace(Result, '', '', [rfReplaceAll]);

  if Length(AWord) < 2 then Exit;

  Index := 0;
  for I := 1 to Length(Result) do
    if Pos(Result[I], VOWEL) > 0 then
    begin
      Index := I;
      Break;
    end;

  if Index = 0 then Exit;

  Prefix := Copy(Result, 1, Index);
  RV := Copy(Result, Index + 1, MAXINT);

  StemStep1(RV);
  StemStep2(RV);
  StemStep3(RV);
  StemStep4(RV);

  Result := Prefix + RV;
end;

function CompareDictItems(List: TStringList; Index1, Index2: Integer): Integer;
begin
  Result := Length(List[Index2]) - Length(List[Index1]);
end;

procedure CreateDict(var Dict: TStringList; const Data: String);
begin
  Dict := TStringList.Create;
  Dict.CommaText := Data;
  Dict.CustomSort(CompareDictItems);
end;

initialization
  CreateDict(DICT_PERFECTIVEGROUND_I, PERFECTIVEGROUND_I);
  CreateDict(DICT_PERFECTIVEGROUND_A, PERFECTIVEGROUND_A);
  CreateDict(DICT_ADJECTIVE, ADJECTIVE);
  CreateDict(DICT_PARTICIPLE_I, PARTICIPLE_I);
  CreateDict(DICT_PARTICIPLE_A, PARTICIPLE_A);
  CreateDict(DICT_VERB_I, VERB_I);
  CreateDict(DICT_VERB_A, VERB_A);
  CreateDict(DICT_NOUN, NOUN);

finalization
  FreeAndNil(DICT_PERFECTIVEGROUND_I);
  FreeAndNil(DICT_PERFECTIVEGROUND_A);
  FreeAndNil(DICT_ADJECTIVE);
  FreeAndNil(DICT_PARTICIPLE_I);
  FreeAndNil(DICT_PARTICIPLE_A);
  FreeAndNil(DICT_VERB_I);
  FreeAndNil(DICT_VERB_A);
  FreeAndNil(DICT_NOUN);


end.


